/*
 * Basic graphics routine by Win32 API.
 *
 * $Id: grwin.c,v 1.1 2000/09/09 22:16:51 muraken Exp $
 */

/*
 * Copyright (C) Kenta Murata, 2000.
 */

#ifndef __GRWIN_C__
#define __GRWIN_C__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <mmsystem.h>

#ifdef _DEBUG
#define debug_message(x) fprintf(stderr, x)
#else /* not _DEBUG */
#define debug_message(x)
#endif /* not _DEBUG */

#define XMAX 640U
#define YMAX 400U

enum {
	BLACK = 0,
	BLUE,
	RED,
	MAGENTA,
	GREEN,
	CYAN,
	YELLOW,
	WHITE
};

COLORREF palette[] = {
	RGB(  0,   0,   0),
	RGB(  0,   0, 255),
	RGB(  0, 255,   0),
	RGB(  0, 255, 255),
	RGB(255,   0,   0),
	RGB(255,   0, 255),
	RGB(255, 255,   0),
	RGB(255, 255, 255)
};

int destroy_flag = 0;
int hitanykey_flag = 0;
ATOM wndclsatom = 0;
HBITMAP bitmap = NULL;
HWND window = NULL;
HINSTANCE instance = NULL;
HANDLE thread = NULL;
HANDLE esc_event = NULL;
HANDLE hitanykey_event = NULL;
HANDLE cw_event = NULL;
RECT client_rect = { 0, 0, XMAX, YMAX };

void
redraw(HDC hdc, LPRECT prc)
{
	HDC hdcmem;
	debug_message("enter redraw()\n");
	hdcmem = CreateCompatibleDC(hdc);
	SelectObject(hdcmem, bitmap);
	BitBlt(hdc, prc->left, prc->top,
	       prc->right - prc->left,
	       prc->bottom - prc->top,
	       hdcmem, prc->left, prc->top, SRCCOPY);
	DeleteDC(hdcmem);
	debug_message("leave redraw()\n");
}

LRESULT CALLBACK
mainwnd_proc(HWND hwnd, UINT msgid, WPARAM wparam, LPARAM lparam)
{
	PAINTSTRUCT ps;
	HDC hdc;
	LPRECT prc;
	switch(msgid) {
	case WM_CREATE:
		return 0;
	case WM_SIZE:
		SetWindowPos(window, NULL, 0, 0,
			     client_rect.right - client_rect.left,
			     client_rect.bottom - client_rect.top,
			     SWP_NOCOPYBITS | SWP_NOMOVE | SWP_NOZORDER);
		return 0;
	case WM_SIZING:
		prc = (LPRECT) lparam;
		switch(wparam) {
		case WMSZ_TOP:
		case WMSZ_TOPLEFT:
		case WMSZ_TOPRIGHT:
		case WMSZ_BOTTOMLEFT:
		case WMSZ_LEFT:
			prc->left = prc->right - (client_rect.right
						  - client_rect.left);
			prc->top = prc->bottom - (client_rect.bottom
						  - client_rect.top);
			break;
		case WMSZ_BOTTOM:
		case WMSZ_BOTTOMRIGHT:
		case WMSZ_RIGHT:
			prc->right = prc->left + (client_rect.right
						  - client_rect.left);
			prc->bottom = prc->top + (client_rect.bottom
						  - client_rect.top);
			break;
		}
		return 0;
	case WM_PAINT:
		debug_message("enter WM_PAINT.\n");
		hdc = BeginPaint(window, &ps);
		redraw(hdc, &ps.rcPaint);
		EndPaint(window, &ps);
		debug_message("leave WM_PAINT.\n");
		return 0;
	case WM_KEYDOWN:
		if(wparam == VK_ESCAPE)
			SetEvent(esc_event);
		if(lparam & (1 << 30))
			return 0;
		if(wparam == VK_F5) {
			hdc = GetDC(window);
			redraw(hdc, &client_rect);
			ReleaseDC(window, hdc);
			return 0;
		}
	case WM_LBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_MBUTTONDOWN:
		debug_message("enter key & mouse check.\n");
		SetEvent(hitanykey_event);
		debug_message("leave key & mouse check.\n");
		return 0;
	case WM_DESTROY:
		debug_message("enter WM_DESTROY.\n");
		PostQuitMessage(0);
		debug_message("leave WM_DESTROY.\n");
		return 0;
	}
	return DefWindowProc(hwnd, msgid, wparam, lparam);
}

void
init_window(void)
{
	WNDCLASSEX wcex;

	debug_message("enter init_window()\n");

	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc = mainwnd_proc;
	wcex.cbClsExtra = wcex.cbWndExtra = 0;
	wcex.hInstance = instance = GetModuleHandle(NULL);
	wcex.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = "grwin";
	wcex.hIconSm = wcex.hIcon;
	wndclsatom = RegisterClassEx(&wcex);
	if(wndclsatom == 0) {
		fprintf(stderr, "Can't register class.\n");
		exit(1);
	}

	AdjustWindowRectEx(
		&client_rect,
		WS_OVERLAPPEDWINDOW,
		FALSE,
		WS_EX_OVERLAPPEDWINDOW);

	window = CreateWindowEx(
		WS_EX_OVERLAPPEDWINDOW,
		"grwin",
		"grwin",
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, CW_USEDEFAULT,
		client_rect.right - client_rect.left,
		client_rect.bottom - client_rect.top,
		NULL, NULL, instance, NULL);

	if(window == NULL) {
		fprintf(stderr, "Can't create window.\n");
		UnregisterClass("grwin", instance);
		exit(1);
	}

	ShowWindow(window, SW_NORMAL);
	UpdateWindow(window);

	debug_message("leave init_window()\n");
}

void
init_bitmap(void)
{
	HDC hdc;
	HDC hdcmem;
	debug_message("enter init_bitmap()\n");

	hdc = GetDC(window);
	hdcmem = CreateCompatibleDC(hdc);
	if(bitmap == NULL) {
		bitmap = CreateCompatibleBitmap(hdc, XMAX, YMAX);
		if(bitmap == NULL) {
			fprintf(stderr, "Can't create bitmap.\n");
			exit(1);
		}
	}
	SelectObject(hdcmem, bitmap);
	FillRect(hdcmem, &client_rect, (HBRUSH)GetStockObject(BLACK_BRUSH));
	DeleteDC(hdcmem);
	ReleaseDC(window, hdc);

	debug_message("leave init_bitmap()\n");
}

DWORD WINAPI
xmain(LPVOID param)
{
	MSG msg;

	debug_message("enter xmain()\n");
	init_window();

	SetEvent(cw_event);

	debug_message("enter message loop.\n");
	while(GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
		if(destroy_flag)
			DestroyWindow(window);
	}
	debug_message("leave message loop.\n");

	hitanykey_flag = 0;
	DeleteObject(bitmap);

	debug_message("leave xmain()\n");

	return msg.wParam;
}

void
hitanykey(void)
{
	char fnbuf[MAX_PATH];
	debug_message("enter hitanykey()\n");
	GetWindowsDirectory(fnbuf, MAX_PATH);
	wsprintf(fnbuf, "\\MEDIA\\CHORD.WAV");
	PlaySound(fnbuf, NULL, SND_FILENAME | SND_SYNC);
	ResetEvent(hitanykey_event);
	WaitForSingleObject(hitanykey_event, INFINITE);
	debug_message("leave hitanykey()\n");
}

void
gr_dot(unsigned int x, unsigned int y, unsigned int color)
{
	HDC hdc;
	HDC hdcmem;
	static unsigned int count = 100;

	if(--count == 0) {
		ResetEvent(esc_event);
		if(WaitForSingleObject(esc_event, 10L) == WAIT_OBJECT_0)
			exit(EXIT_SUCCESS);
		count = 100;
	}
	if((x >= XMAX) || (y >= YMAX))
		return;
	hdc = GetDC(window);
	hdcmem = CreateCompatibleDC(hdc);
	SelectObject(hdcmem, bitmap);
	SetPixel(hdcmem, x, y, palette[color]);
	DeleteDC(hdcmem);
	SetPixel(hdc, x, y, palette[color]);
	ReleaseDC(window, hdc);
}

void
gr_off(void)
{
	DWORD exitcode;

	debug_message("enter gr_off()\n");

	GetExitCodeThread(thread, &exitcode);
	if(exitcode == STILL_ACTIVE) {
		destroy_flag = 1;
		while(exitcode == STILL_ACTIVE)
			GetExitCodeThread(thread, &exitcode);
	}
	if(wndclsatom != 0)
		UnregisterClass("grwin", instance);
	debug_message("leave gr_off()\n");
	CloseHandle(esc_event);
	CloseHandle(hitanykey_event);
	CloseHandle(cw_event);
}

void
gr_on(void)
{
	static int first = 1;

	debug_message("enter gr_on()\n");

	if(first) {
		DWORD thid;

		atexit(gr_off);
		first = 0;

		esc_event = CreateEvent(NULL, FALSE, TRUE, NULL);
		hitanykey_event = CreateEvent(NULL, FALSE, TRUE, NULL);
		cw_event = CreateEvent(NULL, FALSE, FALSE, NULL);
		thread = CreateThread(NULL, 0, xmain, NULL, 0, &thid);
		if(thread == NULL) {
			fprintf(stderr, "Can't create thread.\n");
			exit(1);
		}
		WaitForSingleObject(cw_event, INFINITE);
	}
	init_bitmap();

	debug_message("leave gr_on()\n");
}

#endif /* __GRWIN_C__ */
